# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Apply bug-fixes to the WUFI-XML."""

from PHX.model.hvac.collection import PhxMechanicalSystemCollection
from PHX.model.hvac.heat_pumps import PhxHeatPumpAnnual
from PHX.model.project import PhxProject


def split_cooling_into_multiple_systems(_phx_project: PhxProject):
    """Split the PhxProject's cooling capacity across multiple systems.

    WUFI-Passive v3.x has a bug where any single Ideal-Air-System's cooling capacity may not be > 200 KW.
    This function will split the cooling systems into multiple systems if the total
    cooling load is greater than 200 KW.
    """
    # -- Check is any cooling system capacity is greater than 200 KW

    for phx_variant in _phx_project.variants:
        # ---------------------------------------------------------------------
        # -- Log all the existing cooling systems and their capacities
        total_number_of_cooling_devices = 0
        variant_total_cooling_capacity = 0.0
        variant_total_airflow_rate = 0.0
        total_recirc_cop = 0.0
        total_dehumid_cop = 0.0
        for phx_mech_collection in phx_variant.mech_collections:
            for hp_device in phx_mech_collection.heat_pump_devices:
                # -- If the device doesn't have a cooling system, skip it
                if hp_device.usage_profile.cooling is False:
                    continue

                # -- If the device has a cooling system, add it to the totals
                total_number_of_cooling_devices += 1
                variant_total_cooling_capacity += hp_device.params_cooling.recirculation.capacity
                variant_total_airflow_rate += hp_device.params_cooling.recirculation.flow_rate_m3_hr
                total_recirc_cop += hp_device.params_cooling.recirculation.annual_COP
                total_dehumid_cop += hp_device.params_cooling.dehumidification.annual_COP

        # ---------------------------------------------------------------------
        # -- If the total cooling capacity is less than 200 KW, ignore this variant
        if variant_total_cooling_capacity < 200.0:
            continue

        # ---------------------------------------------------------------------
        # -- Figure out the number of cooling systems needed and the attributes
        print(
            f"Total cooling system capacity = {variant_total_cooling_capacity :.1f} KW. "
            "Splitting the cooling system into multiple systems."
        )

        number_of_new_cooling_systems = int(variant_total_cooling_capacity / 200.0)
        target_number_of_cooling_systems = number_of_new_cooling_systems + 1
        target_number_of_cooling_devices = total_number_of_cooling_devices + number_of_new_cooling_systems
        target_cooling_system_size = variant_total_cooling_capacity / target_number_of_cooling_systems
        target_cooling_system_airflow_rate = variant_total_airflow_rate / target_number_of_cooling_systems
        target_cooling_system_coverage_percentage = 1.0 / target_number_of_cooling_devices
        target_recirc_cop = total_recirc_cop / total_number_of_cooling_devices
        target_dehumid_cop = total_dehumid_cop / total_number_of_cooling_devices
        print(
            f"Splitting {variant_total_cooling_capacity :.1f} KW into "
            f" {target_number_of_cooling_systems} equal-size systems of {target_cooling_system_size} KW."
        )

        # ---------------------------------------------------------------------
        # -- Reset the existing cooling systems
        for phx_mech_collection in phx_variant.mech_collections:
            for hp_device in phx_mech_collection.heat_pump_devices:
                # -- If the device doesn't have a cooling system, skip it
                if hp_device.usage_profile.cooling is False:
                    continue

                # -- If the device has a cooling system, reset it
                hp_device.params_cooling.recirculation.capacity = target_cooling_system_size
                hp_device.params_cooling.recirculation.flow_rate_m3_hr = target_cooling_system_airflow_rate
                hp_device.usage_profile.cooling_percent = target_cooling_system_coverage_percentage
                hp_device.params_cooling.recirculation.annual_COP = target_recirc_cop
                hp_device.params_cooling.dehumidification.annual_COP = target_dehumid_cop

        # ---------------------------------------------------------------------
        # -- Add each of the new cooling systems needed
        for i in range(number_of_new_cooling_systems):
            new_collection = PhxMechanicalSystemCollection()

            new_collection.display_name = "Cooling System [{}]".format(i + 2)

            # -- Build the new Heat Pump for cooling only
            new_cooling_heat_pump = PhxHeatPumpAnnual()
            new_cooling_heat_pump.display_name = "Cooling System [{}]".format(i + 2)
            new_cooling_heat_pump.usage_profile.cooling = True
            new_cooling_heat_pump.params_cooling.recirculation.used = True
            new_cooling_heat_pump.params_cooling.dehumidification.used = True
            new_cooling_heat_pump.params_cooling.recirculation.capacity = target_cooling_system_size
            new_cooling_heat_pump.params_cooling.recirculation.flow_rate_m3_hr = target_cooling_system_airflow_rate
            new_cooling_heat_pump.params_cooling.recirculation.annual_COP = target_recirc_cop
            new_cooling_heat_pump.params_cooling.dehumidification.annual_COP = target_dehumid_cop
            new_cooling_heat_pump.usage_profile.cooling_percent = target_cooling_system_coverage_percentage

            # -- Add the new heat-pump to the mech-collection and to the variant
            new_collection.add_new_mech_device(new_cooling_heat_pump.identifier, new_cooling_heat_pump)
            phx_variant.add_mechanical_collection(new_collection)

    return _phx_project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Occupancy (People) Load."""

from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar


@dataclass
class PhxLoadOccupancy:
    """A PHX Load for the Occupancy (People)."""

    people_per_m2: float = 0.0  # ppl/m2

    @property
    def unique_key(self) -> str:
        return "{}_{:.6f}_".format(self.__class__.__name__, self.people_per_m2)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""DEV SANDBOX: export a specified HBJSON file to a PHPP XL file."""

import pathlib
import sys

import xlwings as xw
from rich import print

from PHX.from_HBJSON import create_project, read_HBJSON_file
from PHX.PHPP import phpp_app
from PHX.xl import xl_app

# --- Input file Path
# -------------------------------------------------------------------------
sys.argv = sys.argv[1:]
# --- Input file Path; optionally provided as first argument
# -------------------------------------------------------------------------
if len(sys.argv) < 1:
    SOURCE_FILE = pathlib.Path(
        "/Users/em/Dropbox/bldgtyp-00/00_PH_Tools/PHX/sample/hbjson/LEVINE_240106.hbjson"
    ).resolve()
else:
    SOURCE_FILE = pathlib.Path(str(sys.argv[0])).resolve()


if __name__ == "__main__":
    # --- Read in an existing HB_JSON and re-build the HB Objects
    # -------------------------------------------------------------------------
    print("[bold green]- " * 50)
    print(f"[bold green]> Reading in the HBJSON file: ./{SOURCE_FILE}[/bold green]")
    hb_json_dict = read_HBJSON_file.read_hb_json_from_file(SOURCE_FILE)
    hb_model = read_HBJSON_file.convert_hbjson_dict_to_hb_model(hb_json_dict)

    # --- Generate the PhxProject file.
    # -------------------------------------------------------------------------
    phx_project = create_project.convert_hb_model_to_PhxProject(hb_model, _group_components=True)

    # --- Connect to open instance of XL, Load the correct PHPP Shape file
    # -------------------------------------------------------------------------
    xl = xl_app.XLConnection(xl_framework=xw, output=print)
    phpp_conn = phpp_app.PHPPConnection(xl)

    try:
        clr = "bold green"
        msg = f"[{clr}]> connected to excel doc: {phpp_conn.xl.wb.name}[/{clr}]"
        xl.output(msg)
    except xl_app.NoActiveExcelRunningError as e:
        raise e

    with phpp_conn.xl.in_silent_mode():
        phpp_conn.xl.unprotect_all_sheets()
        phpp_conn.write_certification_config(phx_project)
        phpp_conn.write_climate_data(phx_project)
        phpp_conn.write_project_constructions(phx_project)
        phpp_conn.write_project_tfa(phx_project)
        phpp_conn.write_project_opaque_surfaces(phx_project)
        phpp_conn.write_project_thermal_bridges(phx_project)
        phpp_conn.write_project_window_components(phx_project)
        phpp_conn.write_project_window_surfaces(phx_project)
        phpp_conn.write_project_window_shading(phx_project)
        phpp_conn.write_project_ventilation_components(phx_project)
        phpp_conn.write_project_ventilators(phx_project)
        phpp_conn.write_project_spaces(phx_project)
        phpp_conn.write_project_ventilation_type(phx_project)
        phpp_conn.write_project_airtightness(phx_project)
        phpp_conn.write_project_volume(phx_project)
        phpp_conn.write_project_hot_water(phx_project)
        phpp_conn.write_project_res_elec_appliances(phx_project)
        phpp_conn.write_non_res_utilization_profiles(phx_project)
        phpp_conn.write_non_res_space_lighting(phx_project)
        phpp_conn.write_non_res_IHG(phx_project)

        # TODO: add custom any-range writer (User-Determined)

        phpp_conn.activate_variant_assemblies()
        phpp_conn.activate_variant_windows()
        phpp_conn.activate_variant_ventilation()
        phpp_conn.activate_variant_additional_vent()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Functions to allow the merging of HB-Faces which are used to simplify HB-Models."""

import logging
from typing import List, Sequence, Tuple, TypeVar

try:
    from honeybee.aperture import Aperture
    from honeybee.face import Face
    from honeybee.shade import Shade
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from ladybug_geometry.geometry2d.polygon import Polygon2D
    from ladybug_geometry.geometry3d.face import Face3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_geometry:\n\t{}".format(e))

try:
    from honeybee_ph_utils import polygon2d_tools
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph_utils:\n\t{}".format(e))


logger = logging.getLogger()

TFaceOrShade = TypeVar("TFaceOrShade", Face, Shade)


def _add_sub_face(_face: Face, _aperture: Aperture):
    """Add an HB-sub-face (either HB-Aperture or HB-Door) to a parent Face.

    NOTE: this method is copied from honeybee's Grasshopper component "HB Add Sub-face"
    """
    if isinstance(_aperture, Aperture):  # the sub-face is an Aperture
        _face.add_aperture(_aperture)
    else:  # the sub-face is a Door
        _face.add_door(_aperture)

    return _face


def _check_and_add_sub_face(
    _face: Face,
    _apertures: List[Aperture],
    _tolerance: float,
    _angle_tolerance: float,
) -> None:
    """Check whether a HB-sub-face is valid for an HB-face and, if so, add it.

    NOTE: this method is copied from honeybee's Grasshopper component "HB Add Sub-face"
    """
    for aperture in _apertures:
        if _face.geometry.is_sub_face(aperture.geometry, _tolerance, _angle_tolerance):
            _add_sub_face(_face, aperture)


def _create_new_HB_Face(_face3D: Face3D, _ref_face: Face) -> Face:
    """Create a new HB-Face using a Face3D and a reference HB-Face."""
    new_face = Face(
        identifier=_ref_face.identifier,
        geometry=_face3D,
        type=_ref_face.type,
        boundary_condition=_ref_face.boundary_condition,
    )
    new_face.display_name = _ref_face.display_name
    new_face._user_data = None if _ref_face.user_data is None else _ref_face.user_data.copy()
    new_face._properties._duplicate_extension_attr(_ref_face._properties)

    return new_face


def _create_new_HB_Shade(_face3D: Face3D, _ref_face: Shade) -> Shade:
    """Create a new HB-Shade using a Face3D and a reference HB-Shade."""
    new_face = Shade(
        identifier=_ref_face.identifier,
        geometry=_face3D,
        is_detached=True,
    )
    new_face.display_name = _ref_face.display_name
    new_face._user_data = None if _ref_face.user_data is None else _ref_face.user_data.copy()
    new_face._properties._duplicate_extension_attr(_ref_face._properties)

    return new_face


def _create_new_Face3D(_poly2D: Polygon2D, _base_plane: Plane, _ref_face: TFaceOrShade) -> Face3D:
    """Return a new Face3D based on a Polygon2D and a reference HB-Face."""
    return Face3D(
        boundary=tuple(_base_plane.xy_to_xyz(v) for v in _poly2D.vertices),
        plane=_ref_face.geometry.plane,
    )


def find_parent_and_child_polygons(
    _polygons: List[Polygon2D],
) -> Tuple[List[Polygon2D], List[Polygon2D]]:
    """Return a Tuple of any parent (container) and child (contained) polygons from a list of polygons."""

    # Initialize empty lists for parent and child surfaces
    parent_polygon = []
    child_polygons = []

    # Compare each surface with all other surfaces
    for i, _polygon in enumerate(_polygons):
        is_inside_any = False

        for j, other_polygon in enumerate(_polygons):
            if i == j:
                continue

            if _polygon.is_polygon_inside(other_polygon):
                is_inside_any = True
                break

        # If the surface is inside any other surface, add it to child_surfaces
        # Otherwise, it is a parent surface
        if is_inside_any:
            parent_polygon.append(_polygon)
        else:
            child_polygons.append(_polygon)

    return parent_polygon, child_polygons


def face3Ds(_faces: Sequence[TFaceOrShade]) -> List[Face3D]:
    """Return a list of LBT-Face3Ds from a list of HB-Faces or HB-Shades."""
    return [f.geometry for f in _faces]


def merge_hb_faces(_faces: List[Face], _tolerance: float, _angle_tolerance_degrees: float) -> List[Face]:
    """Merge a group of HB-Faces into the fewest number of faces possible."""

    if not _faces:
        logger.debug("No faces in group. Skipping merge.")
        return []

    if len(_faces) == 1:
        logger.debug(f"Single face: {_faces[0].display_name} in group. Skipping merge.")
        return _faces

    # -------------------------------------------------------------------------
    # -- Before anything else, preserve any the Apertures for adding back in later
    apertures = []
    for f in _faces:
        apertures.extend([ap.duplicate() for ap in getattr(f, "apertures", [])])

    # -------------------------------------------------------------------------
    # -- Merge the HB-Faces' Polygons together
    merged_polygon2Ds = polygon2d_tools.merge_lbt_face_polygons(face3Ds(_faces), _tolerance)

    # -------------------------------------------------------------------------
    # -- Create new LBT-Face3D, and HB-Faces from the Polygon2Ds
    ref_face = _faces[0]
    ref_plane = ref_face.geometry.plane
    faces = []
    if len(merged_polygon2Ds) == 1:
        # -- Create new faces for the merged Polygon2Ds
        face3ds = (_create_new_Face3D(p, ref_plane, ref_face) for p in merged_polygon2Ds)
        faces = [_create_new_HB_Face(f3d, ref_face) for f3d in face3ds]
    elif len(merged_polygon2Ds) > 1:
        # -- It may mean that there are 'holes' in a surface? So try and find
        # -- the parent and any child surfaces.

        parent_polygon, child_polygons = find_parent_and_child_polygons(merged_polygon2Ds)

        # -- Check the results
        if len(parent_polygon) != 1:
            # -- Something went wrong, give up.
            _faces.append(ref_face)
            return _faces

        # -- If only 1 parent, lets make some Face3Ds and Faces
        parent_face_3d = [_create_new_Face3D(p, ref_plane, ref_face) for p in parent_polygon]
        child_face_3ds = [_create_new_Face3D(p, ref_plane, ref_face) for p in child_polygons]
        face_3ds = [Face3D.from_punched_geometry(parent_face_3d[0], child_face_3ds)]
        faces = [_create_new_HB_Face(f3d, ref_face) for f3d in face_3ds]

    # -------------------------------------------------------------------------
    # -- Add the apertures back in
    faces_with_apertures_: List[Face] = []
    for _face in faces:
        _check_and_add_sub_face(_face, apertures, _tolerance, _angle_tolerance_degrees)
        faces_with_apertures_.append(_face)

    return faces_with_apertures_


def merge_hb_shades(_faces: List[Shade], _tolerance: float, _angle_tolerance_degrees: float) -> List[Shade]:
    """Merge a group of HB-Shades into the fewest number of shades possible."""
    if not _faces:
        return []

    if len(_faces) == 1:
        return list(_faces)

    # -------------------------------------------------------------------------
    # -- Merge the HB-Face's Polygons together
    merged_polygon2Ds = polygon2d_tools.merge_lbt_face_polygons(face3Ds(_faces), _tolerance)

    # -------------------------------------------------------------------------
    # -- Create new LBT-Face3D, and HB-Faces from the Polygon2Ds
    ref_face = _faces[0]
    ref_plane = ref_face.geometry.plane
    hb_shades_ = []
    if len(merged_polygon2Ds) == 1:
        # -- Create new faces for the merged Polygon2Ds
        face3ds = (_create_new_Face3D(p, ref_plane, ref_face) for p in merged_polygon2Ds)
        hb_shades_ = [_create_new_HB_Shade(f3d, ref_face) for f3d in face3ds]
    elif len(merged_polygon2Ds) > 1:
        # -- It may mean that there are 'holes' in a surface? So try and find
        # -- the parent and any child surfaces.

        parent_polygon, child_polygons = find_parent_and_child_polygons(merged_polygon2Ds)

        # -- Check the results
        if len(parent_polygon) != 1:
            # -- Something went wrong, give up.
            _faces.append(ref_face)
            return _faces

        # -- If only 1 parent, lets make some Face3Ds and Faces
        parent_face_3d = [_create_new_Face3D(p, ref_plane, ref_face) for p in parent_polygon]
        child_face_3ds = [_create_new_Face3D(p, ref_plane, ref_face) for p in child_polygons]
        face_3ds = Face3D.from_punched_geometry(parent_face_3d[0], child_face_3ds)
        hb_shades_ = [_create_new_HB_Shade(f3d, ref_face) for f3d in face_3ds]

    return hb_shades_
